#!/usr/bin/perl

use strict;

######This program accepts an mps file of a metabolic network, a nutrient file,
######and a reaction to gene relations file as input and will determine the 
######crypt/pseudogenes that are responsible for a phenotypic behavior.

my $mps="Pathway.mps";
my $env="environment.txt";
my $rxgene="Reaction_gene_association.txt";

open (RXG,$rxgene);      #read the reaction_gene realationships and insert in an array
my @rel=<RXG>;                         
shift @rel;                                     
my $rlar=\@rel;
close RXG;

open (PATH,$mps);        #read the initial mps file
my @lines=<PATH>;
my $liar=\@lines;
close PATH;

open (MET,$env);         #read the environment file
my @temp=<MET>;
chomp @temp;
shift @temp;
my $MeioNature=shift @temp;   #Nature of the meiotrophic process
$MeioNature=~ s/^\s+//;
$MeioNature=~ s/\s+$//;
shift @temp;
my $khat=shift @temp;
my $meio=substr ($khat,0,10);   #Meiotrophic metabolite
$meio=~ s/^\s+//;
$meio=~ s/\s+$//;
$khat=substr ($khat,10);        #processes that import the meiotrophic compound
$khat=~ s/^\s+//;
$khat=~ s/\s+$//;
my @mrxns=split (/_+/,$khat);
my $mxar=\@mrxns;

my $wildtype="no";

my %mets=();
foreach (@temp){                #List of alternate metabolites
    if (($_ eq "") or ($_=~/\#/)){
	next;
    }else{
	my $metabolite=substr($_,0,10);
	$metabolite=~ s/^\s+//;
	$metabolite=~ s/\s+$//;
	my $rxns=substr($_,10);
	if ($metabolite eq "WT"){
	    $wildtype="yes";
	    $rxns="none";
	}
	my @holder=split (/_+/,$rxns);
	foreach (@holder){
	    push(@{$mets{$metabolite}},$_);
	}
    }
}

my @metabolites=keys (%mets);
@metabolites=sort(@metabolites);

my @genes=();
my $gear=\@genes;
&genes($rlar,$gear);     #get a list of all genes associated with the models
@genes=sort(@genes);

my %grx=();
my $grhr=\%grx;

foreach (@genes){
    my $holder=$_;
    my @rxns=();
    my $offs=\@rxns;
    &rxns($holder,$rlar,$offs);    #get a list of all reactions for each gene to be turned off
    my $count=$#rxns;
    unless ($count==-1){
	push(@{$grx{$holder}},@rxns);
    }
}


my %criticalgenes=();
my $crhr=\%criticalgenes;
my $tfile="temp.mps";

foreach (@metabolites){
    my $holder=$_;
    my @rxns=@mrxns;
    my $offrx=\@rxns;
    my $onrx=\@{$mets{$_}};
    foreach (@metabolites){
	unless ($_ eq $holder){
	    push (@rxns,@{$mets{$_}});        #import of meitrophic compound and all but one of the metabolites to be turned off
	}
    }
    unless ($holder eq "WT"){
	&makemps ($offrx,$onrx,$liar);
    }else{
	my $v="NORXNS";
	&makemps ($offrx,$v,$liar);
    }
    my %zemet=();
    my $zmhr=\%zemet;
    my $ans=&glpflux($tfile,$zmhr);          
    if (($ans eq "INF") or ($ans eq "UB")){
	print "$holder\t Infeasible or unbound problem\n";
	exit;
    }
    &sgko($tfile,$crhr,$gear,$grhr,$zmhr);   #knockout all genes under above condition and identify the critical genes
}

my @mcg=keys (%criticalgenes);    #list of all the genes that are critical for all alternate metabolites
@mcg=sort (@mcg);

my %meiocriticalgenes=();
my $mehr=\%meiocriticalgenes;

my $v="NORXNS";
if ($wildtype eq "yes"){           #if the initial environment is labeled WT when meiotrophic import is on, no reaction will be turned off
    &makemps ($v,$mxar,$liar);
}else{
    if ($MeioNature eq "BS"){
	foreach (@metabolites){
	    my @offrx=();
	    my @onrx=();
	    my $holder=$_;
	    foreach (@metabolites){
		if ($_ eq $holder){
		    push (@onrx,@{$mets{$_}});
		}else{
		    push (@offrx,@{$mets{$_}});
		}
	    }
	    push (@onrx,@mrxns);
	    my $bsar=\@onrx;
	    my $ncar=\@offrx;
	    &makemps ($ncar,$bsar,$liar);
	    my %zemet=();
	    my $zmhr=\%zemet;
	    my $ans=&glpflux($tfile,$zmhr);

	    if (($ans eq "INF") or ($ans eq "UB")){
		print "Infeasible or unbound problem\n";
		exit;
	    }
	    &sgko($tfile,$mehr,$gear,$grhr,$zmhr);
	}
    }elsif ($MeioNature eq "NC"){
	my @offrx=();
	my $ncar=\@offrx;
	foreach (@metabolites){
	    push (@offrx,@{$mets{$_}});
	}
	&makemps ($ncar,$mxar,$liar);
	my %zemet=();
	my $zmhr=\%zemet;
	my $ans=&glpflux($tfile,$zmhr);

	if (($ans eq "INF") or ($ans eq "UB")){
	    print "Infeasible or unbound problem\n";
	    exit;
	}
	&sgko($tfile,$mehr,$gear,$grhr,$zmhr);
    }
}

my @ocg=keys (%meiocriticalgenes);   #genes that are critical when meiotrophic import is turned on
@ocg=sort (@ocg);

print "\nThe meiotrophic genes for $meio are:\n"; 

if ($MeioNature eq "BS"){           #for biosynthetic conditions meiotrophic genes are found by @mcg-@ocg
    foreach (@mcg){
	unless (exists $meiocriticalgenes{$_}){
	    print "$_\n";
	}
    }
}elsif ($MeioNature eq "NC"){       #for nutrition consumption meiotrophic genes are found by @ocg-@mcg
    foreach (@ocg){
	unless (exists $criticalgenes{$_}){
	    print "$_\n";
	}
    }
}

system ("rm -f $tfile");
exit;

###################
sub sgko{
    my ($tfile,$crhr,$gear,$grhr,$zmhr)=@_;
    my @genes=@$gear;
    open (PATH,$tfile);
    my @lines=<PATH>;
    close PATH;
    my $ltar=\@lines;
    foreach (@genes){
	unless (exists ($$crhr{$_})){
	    if (exists ($$grhr{$_})){
		my @offrx=@{$$grhr{$_}};
		my $offr=\@offrx;
		my $v="NORXNS";
		my $test="no";
		foreach (@offrx){
		    unless (exists $$zmhr{$_}){
			$test="yes";
		    }
		}
		if ($test eq "yes"){
		    &makemps($offr,$v,$ltar);
		    my $ofval=&objfnval;
		    if (abs($ofval)<1.e-3){
			$$crhr{$_}="yes";
		    }
		}
	    }
	}
    }
}


##################
sub rxns{
    my ($holder,$rlar,$offs)=@_;
    my @lines=@$rlar;
    chomp @lines;
    my $pre="_".$holder;
    my $post=$holder."_";
    foreach (@lines){
	my $rxname=substr ($_,0,10);
	$rxname=~ s/^\s+//;
	$rxname=~ s/\s+$//;
	my $rest=substr ($_,10);
	$rest=~ s/^\s+//;
	$rest=~ s/\s+$//;
	if ($rest=~/$holder/){
	    unless (($_=~/$pre/) or ($_=~/$post/)){
		push (@$offs,$rxname);
	    }
	}
    }
}
	

##################
sub makemps{
    my ($offrx,$onrx,$liar)=@_;
    my @rxns=();
    my @ons=();
    unless ($onrx eq "NORXNS"){
	@ons=@$onrx;
    }
    unless ($offrx eq "NORXNS"){
	@rxns=@$offrx;
    }
    my @lines=@$liar;
    my $c=0;
    open (FILL,">temp.mps");
    until ($lines[$c-1]=~/BOUNDS/){
	print FILL $lines[$c];
	$c+=1;
    }
    until ($lines[$c]=~/ENDATA/){
	my $rxname=substr($lines[$c],14,10);
	$rxname=~ s/^\s+//;
	$rxname=~ s/\s+$//;
	my $test="no";
	foreach (@rxns){
	    if ($_ eq $rxname){
		$test="yes";
	    }
	}
	foreach (@ons){
	    if ($_ eq $rxname){
		$test="yes";
	    }
	}
	unless ($test eq "yes"){
	    print FILL $lines[$c];
	}
	$c+=1;
    }
    foreach (@rxns){
	my $phrase=" FX BND1      ";
	my $blank=" "x(10-length($_));
	my $value="0.000";
	print FILL "$phrase$_$blank$value\n";
    }
    foreach (@ons){
	my $phrase=" UP BND1      ";
	my $blank=" "x(10-length($_));
	my $value="10.00";
	print FILL "$phrase$_$blank$value\n";
    }
    print FILL $lines[$c];
    close FILL;
}
    
####################

sub genes{
    my ($rlar,$gear)=@_;
    my @array=@$rlar;
    chomp @array;
    my %hash=();
    foreach (@array){
	my @holder=();
	my $genes=substr ($_,10);
	unless ($genes=~/\)_\(/){
	    @holder=split (/and+/,$genes);
	    foreach (@holder){
		$_=~ s/^\s+//;
		$_=~ s/\s+$//;
		$_=~ s/\(//g;
		$_=~ s/\)//g;
#		$_=~ s/ //g;
	    }
	}else{
	    my $ind=index ($genes,")_(");
	    my $early=substr($genes,0,$ind);
	    my $late=substr($genes,$ind+2);
	    my @one=split (/and+/,$early);
	    my @two=split (/and+/,$late);
	    foreach (@one){
		$_=~ s/^\s+//;
		$_=~ s/\s+$//;
		$_=~ s/\(//g;
		$_=~ s/\)//g;
	    }
	    foreach (@two){
		$_=~ s/^\s+//;
		$_=~ s/\s+$//;
		$_=~ s/\(//g;
		$_=~ s/\)//g;
	    }
	    push (@one,@two);
	    @one=sort (@one);
	    @holder=@one;
	}

	foreach (@holder){
	    my @fini=split (/_+/);
	    foreach (@fini){
		my $postand=$_;
		unless (exists $hash{$postand}){
		    $hash{$postand}=1;
		}
	    }
	}

    }
    my @results=keys (%hash);
    @$gear=sort(@results);
}

#################
sub glpflux {
    my $ash="ashghal";
    my $results="results";
    my ($wild,$wthr)=@_;
    system ("glpsol $wild -o $results >$ash");
    open (JUNK,$ash);
    my @jj=<JUNK>;
    close (JUNK);
    system ("rm $ash");
    my $emtehan="no";
    my $ubtest="no";
    foreach (@jj){
	if ($_=~/OPTIMAL SOLUTION FOUND/){
	    $emtehan="yes";
	}elsif($_=~/PROBLEM HAS UNBOUNDED SOLUTION/){
	    $ubtest="UB";
	}
    }
    if ($emtehan eq "yes"){
	open (FIRST, $results);
	my @glp=<FIRST>;
	close FIRST;
	my $c3=0;
	until ($glp[$c3]=~/Column name/){
	    $c3+=1;
	}
	$c3=$c3+2;
	until ($glp[$c3+1]=~/Karush/){
	    my $name=substr ($glp[$c3],7,12);
	    my $flux=substr ($glp[$c3],23,14);
	    $name=~ s/^\s+//;
	    $name=~ s/\s+$//;
	    $flux=~ s/^\s+//;
	    $flux=~ s/\s+$//;
	    if ($flux==0){
		$$wthr{$name}=$flux;
	    }
	    $c3+=1;
	}
    } else {
	if ($ubtest eq "no"){
	    my $vvv="INF";
	    return $vvv;
	}else{
	    return $ubtest;
	}
    }
    system ("rm results");
	
}

###################
sub objfnval{
    my $ash="ashghal";
    my $results="results";
    my $mps="temp.mps";
    my $c=9;
    system ("glpsol $mps -o $results >$ash");
    open (JUNK,$ash);
    my @jj=<JUNK>;
    close (JUNK);
    system ("rm $ash");
    my $emtehan="no";
    my $ubtest="no";
    my $flux="";
    foreach (@jj){
	if ($_=~/OPTIMAL SOLUTION FOUND/){
	    $emtehan="yes";
	}elsif($_=~/PROBLEM HAS UNBOUNDED SOLUTION/){
	    $ubtest="UB";
	}
    }
    if ($emtehan eq "yes"){
	open (FIRST, $results);
	my @glp=<FIRST>;
	close FIRST;
	my $costcount=0;
	my $c=0;
	foreach(@glp){
	    if ($_=~/COST/){
		$costcount+=1;
	    }
	    if ($costcount==2){
		$flux=substr ($glp[$c],23,14);
		$flux=~ s/^\s+//;
		$flux=~ s/\s+$//;
		$costcount+=1;
	    }
	    $c+=1;
	}
    } else {
	if ($ubtest eq "no"){
	    my $vvv="INF";
	    return $vvv;
	}else{
	    return $ubtest;
	}
    }
    system ("rm results");    
    return $flux;
}





